'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var estreeWalker = require('estree-walker');
var acorn = require('acorn');
var MagicString = _interopDefault(require('magic-string'));

var patterns = [
  {
    name: 'isUndefined',
    func: "function (x) { return typeof x === \"undefined\" }",
    tests: [
      {
        test: function test (node) {
          return node.type === 'BinaryExpression' &&
            (node.operator === '===' || node.operator === '==') &&
            node.left.type === 'UnaryExpression' &&
            node.left.operator === 'typeof' &&
            node.right.type === 'Literal' &&
            node.right.value === 'undefined'
        },
        getOffsets: function getOffsets (node) {
          return [node.start, node.left.argument.start, node.left.argument.end, node.end]
        }
      },
      {
        test: function test$1 (node) {
          return node.type === 'BinaryExpression' &&
            (node.operator === '===' || node.operator === '==') &&
            node.right.type === 'UnaryExpression' &&
            node.right.operator === 'typeof' &&
            node.left.type === 'Literal' &&
            node.left.value === 'undefined'
        },
        getOffsets: function getOffsets$1 (node) {
          return [node.start, node.right.argument.start, node.right.argument.end, node.end]
        }
      },
      {
        test: function test$2 (node) {
          return node.type === 'BinaryExpression' &&
            node.operator === '===' &&
            node.right.type === 'Identifier' &&
            node.right.name === 'undefined'
        },
        getOffsets: function getOffsets$2 (node) {
          return [node.start, node.left.start, node.left.end, node.end]
        }
      },
      {
        test: function test$3 (node) {
          return node.type === 'BinaryExpression' &&
            node.operator === '===' &&
            node.left.type === 'Identifier' &&
            node.left.name === 'undefined'
        },
        getOffsets: function getOffsets$3 (node) {
          return [node.start, node.right.start, node.right.end, node.end]
        }
      },
      {
        test: function test$4 (node) {
          return node.type === 'BinaryExpression' &&
            (node.operator === '!==' || node.operator === '!=') &&
            node.left.type === 'UnaryExpression' &&
            node.left.operator === 'typeof' &&
            node.right.type === 'Literal' &&
            node.right.value === 'undefined'
        },
        getOffsets: function getOffsets$4 (node) {
          return [node.start, node.left.argument.start, node.left.argument.end, node.end]
        },
        negated: true
      },
      {
        test: function test$5 (node) {
          return node.type === 'BinaryExpression' &&
            (node.operator === '!==' || node.operator === '!=') &&
            node.right.type === 'UnaryExpression' &&
            node.right.operator === 'typeof' &&
            node.left.type === 'Literal' &&
            node.left.value === 'undefined'
        },
        getOffsets: function getOffsets$5 (node) {
          return [node.start, node.right.argument.start, node.right.argument.end, node.end]
        },
        negated: true
      },
      {
        test: function test$6 (node) {
          return node.type === 'BinaryExpression' &&
            node.operator === '!==' &&
            node.right.type === 'Identifier' &&
            node.right.name === 'undefined'
        },
        getOffsets: function getOffsets$6 (node) {
          return [node.start, node.left.start, node.left.end, node.end]
        },
        negated: true
      },
      {
        test: function test$7 (node) {
          return node.type === 'BinaryExpression' &&
            node.operator === '!==' &&
            node.left.type === 'Identifier' &&
            node.left.name === 'undefined'
        },
        getOffsets: function getOffsets$7 (node) {
          return [node.start, node.right.start, node.right.end, node.end]
        },
        negated: true
      }
    ]
  },
  {
    name: 'isNull',
    func: "function (x) { return x === null }",
    tests: [
      {
        test: function test$8 (node) {
          return node.type === 'BinaryExpression' &&
            node.operator === '===' &&
            node.right.type === 'Literal' &&
            node.right.raw === 'null'
        },
        getOffsets: function getOffsets$8 (node) {
          return [node.start, node.left.start, node.left.end, node.end]
        }
      },
      {
        test: function test$9 (node) {
          return node.type === 'BinaryExpression' &&
            node.operator === '===' &&
            node.left.type === 'Literal' &&
            node.left.raw === 'null'
        },
        getOffsets: function getOffsets$9 (node) {
          return [node.start, node.right.start, node.right.end, node.end]
        }
      },
      {
        test: function test$10 (node) {
          return node.type === 'BinaryExpression' &&
            node.operator === '!==' &&
            node.right.type === 'Literal' &&
            node.right.raw === 'null'
        },
        getOffsets: function getOffsets$10 (node) {
          return [node.start, node.left.start, node.left.end, node.end]
        },
        negated: true
      },
      {
        test: function test$11 (node) {
          return node.type === 'BinaryExpression' &&
            node.operator === '!==' &&
            node.left.type === 'Literal' &&
            node.left.raw === 'null'
        },
        getOffsets: function getOffsets$11 (node) {
          return [node.start, node.right.start, node.right.end, node.end]
        },
        negated: true
      }
    ]
  }
];

['function', 'number', 'boolean', 'string', 'object'].forEach(function (type) {
  patterns.push({
    name: ("is" + (type.charAt(0).toUpperCase() + type.substring(1))),
    func: ("function (x) { return typeof x === \"" + type + "\" }"),
    tests: [
      {
        test: function test (node) {
          return node.type === 'BinaryExpression' &&
            (node.operator === '===' || node.operator === '==') &&
            node.left.type === 'UnaryExpression' &&
            node.left.operator === 'typeof' &&
            node.right.type === 'Literal' &&
            node.right.value === type
        },
        getOffsets: function getOffsets (node) {
          return [node.start, node.left.argument.start, node.left.argument.end, node.end]
        }
      },
      {
        test: function test$1 (node) {
          return node.type === 'BinaryExpression' &&
            (node.operator === '===' || node.operator === '==') &&
            node.right.type === 'UnaryExpression' &&
            node.right.operator === 'typeof' &&
            node.left.type === 'Literal' &&
            node.left.value === type
        },
        getOffsets: function getOffsets$1 (node) {
          return [node.start, node.right.argument.start, node.right.argument.end, node.end]
        }
      },
      {
        test: function test$2 (node) {
          return node.type === 'BinaryExpression' &&
            (node.operator === '!==' || node.operator === '!=') &&
            node.left.type === 'UnaryExpression' &&
            node.left.operator === 'typeof' &&
            node.right.type === 'Literal' &&
            node.right.value === type
        },
        getOffsets: function getOffsets$2 (node) {
          return [node.start, node.left.argument.start, node.left.argument.end, node.end]
        },
        negated: true
      },
      {
        test: function test$3 (node) {
          return node.type === 'BinaryExpression' &&
            (node.operator === '!==' || node.operator === '!=') &&
            node.right.type === 'UnaryExpression' &&
            node.right.operator === 'typeof' &&
            node.left.type === 'Literal' &&
            node.left.value === type
        },
        getOffsets: function getOffsets$3 (node) {
          return [node.start, node.right.argument.start, node.right.argument.end, node.end]
        },
        negated: true
      }
    ]
  });
});

function tryParse (code, id) {
  try {
    return acorn.parse(code, {
      ecmaVersion: 6,
      sourceType: 'module'
    })
  } catch (err) {
    console.warn(("rollup-plugin-inline-functions: failed to parse " + id + ". Consider restricting the plugin to particular files via options.include"));
  }
}

function namespace (name) {
  return ("__inject_" + name + "__")
}

function inlineTypeof (options) {
  options = options || {};
  var sourceMap = options.sourceMap !== false;

  var patternImporteeNames = patterns.map(function (pattern) { return namespace(pattern.name); });
  var patternImporteeNamesToPatterns = {};
  patterns.forEach(function (pattern) {
    patternImporteeNamesToPatterns[namespace(pattern.name)] = pattern;
  });

  return {
    name: 'inline-typeof',

    resolveId: function resolveId (importee) {
      if (patternImporteeNames.indexOf(importee) !== -1) {
        return importee
      }
      return null
    },

    load: function load (id) {
      if (patternImporteeNames.indexOf(id) !== -1) {
        var pattern = patternImporteeNamesToPatterns[id];
        var func = pattern.func;
        return ("export default " + (func.replace(/^function/, ("function " + (pattern.name)))))
      }
      return null
    },

    transform: function transform (code, id) {
      if (patternImporteeNames.indexOf(id) !== -1) {
        return null
      }
      var ast = tryParse(code, id);
      if (!ast) { return null }

      var magicString = new MagicString(code);

      var newImports = {};
      var newImportHashes = {};

      estreeWalker.walk(ast, {
        enter: function enter (node) {
          if (sourceMap) {
            magicString.addSourcemapLocation(node.start);
            magicString.addSourcemapLocation(node.end);
          }

          patterns.forEach(function (pattern) {
            pattern.tests.forEach(function (test) {
              if (test.test(node)) {
                var importName = namespace(pattern.name);
                var hash;
                if (newImports[importName]) {
                  hash = newImportHashes[importName];
                } else {
                  hash = importName + "_" + (Math.round(1000000000 * Math.random()));
                  newImports[importName] = "import " + hash + " from '" + importName + "';";
                  newImportHashes[importName] = hash;
                }
                var ref = test.getOffsets(node);
                var leftStart = ref[0];
                var leftEnd = ref[1];
                var rightStart = ref[2];
                var rightEnd = ref[3];
                var prepend = "" + (test.negated ? '!' : '') + hash + "(";
                var append = ")";
                if (leftStart === leftEnd) {
                  magicString.insertLeft(leftStart, prepend);
                } else {
                  magicString.overwrite(leftStart, leftEnd, prepend);
                }
                if (rightStart === rightEnd) {
                  magicString.insertRight(rightStart, append);
                } else {
                  magicString.overwrite(rightStart, rightEnd, append);
                }
              }
            });
          });
        },
        leave: function leave (node) {
        }
      });

      var keys = Object.keys(newImports);
      if (!keys.length) { return null }

      var importBlock = keys.map(function (key) { return newImports[ key ]; }).join('\n\n');
      magicString.prepend(importBlock + '\n\n');

      return {
        code: magicString.toString(),
        map: sourceMap ? magicString.generateMap() : null
      }
    }
  }
}

module.exports = inlineTypeof;
